unit Pendula;

// C Code from billp245 obtained from :
//
//  https://www.physicsforums.com/threads/double-pendulum-in-c-trace-motion-real-time-gif-output.818983/
//
// There were no copyright messages available from there.
//
//
// Converted to Delphi by Blue Hell

{

   COPYRIGHT 2020 Blue Hell / Jan Punter (for the modifications only)

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation;

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  For all listed email addresses :

    _dot. to be substituted by a dot      '.'
    2@t2  to be substituted by an at sign '@'


  Blue Hell is a trade mark owned by

    Jan Punter
    https://www.bluehell.nl/
    jan2@t2mail_dot_bluehell_dot_nl
}


interface

uses

  System.Math,

  KnobsUtils, KnobsConversions;

type

  TPendulum = class
  strict private
  const
    INCREMENT = 0.0175;
  private
    FSpeed : TSignal;
    FGMod  : TSignal;
    FNoise : TSignal;
    FL1    : TSignal;
    FL2    : TSignal;
    FM1    : TSignal;
    FM2    : TSignal;
    FTh1   : TSignal;
    FTh2   : TSignal;
    FT1d   : TSignal;
    FT2d   : TSignal;
  private
    function    Jitter: TSignal;
    function    th1_xder( t1d: TSignal): TSignal;
    function    th1_yder( t2d: TSignal): TSignal;
    function    th2_xder( t1d, th1, t2d, th2, l1, l2, m1, m2: TSignal): TSignal;
    function    th2_yder( t1d, th1, t2d, th2, l1, l2, m1, m2: TSignal): TSignal;
  public
    constructor Create( l1, l2, m1, m2: TSignal);
    procedure   Reset;
    procedure   Step( out x1, y1, x2, y2, ph1, ph2: TSignal; fetch_only: Boolean);
  public
    property    L1    : TSignal read FL1    write FL1;
    property    L2    : TSignal read FL2    write FL2;
    property    M1    : TSignal read FM1    write FM1;
    property    M2    : TSignal read FM2    write FM2;
    property    Speed : TSignal read FSpeed write FSpeed;
    property    GMod  : TSignal read FGMod  write FGMod;
    property    Noise : TSignal read FNoise write FNoise;
  end;


implementation


{ ========
  TPendulum = class
  private
    FSpeed : TSignal;
    FGMod  : TSignal;
    FNoise : TSignal;
    FL1    : TSignal;
    FL2    : TSignal;
    FM1    : TSignal;
    FM2    : TSignal;
    FTh1   : TSignal;
    FTh2   : TSignal;
    FT1d   : TSignal;
    FT2d   : TSignal;
  public
    property    L1    : TSignal read FL1    write FL1;
    property    L2    : TSignal read FL2    write FL2;
    property    M1    : TSignal read FM1    write FM1;
    property    M2    : TSignal read FM2    write FM2;
    property    Speed : TSignal read FSpeed write FSpeed;
    property    GMod  : TSignal read FGMod  write FGMod;
    property    Noise : TSignal read FNoise write FNoise;
  private
}

    function    TPendulum.Jitter: TSignal;
    const
      FIVEPERC = 5.0 / 100.0;     // +/- 5% so 10% in total
    begin
      Result := 1.0 + FIVEPERC * FNoise * ( Random - 2.0 * Random);
    end;


    function    TPendulum.th1_xder( t1d: TSignal): TSignal;
    begin
      Result := t1d * INCREMENT * FSpeed * Jitter;
    end;


    function    TPendulum.th1_yder( t2d: TSignal): TSignal;
    begin
      Result := t2d * INCREMENT * FSpeed * Jitter;
    end;


    function    TPendulum.th2_xder( t1d, th1, t2d, th2, l1, l2, m1, m2: TSignal): TSignal;
    begin
      Result :=
        INCREMENT * FSpeed * Jitter *
        (
          (( STANDARD_GRAVITY * FGMod) / l1) * ( ( m2 / ( m1 + m2)) * Sin( th2) * Cos( th1 - th2) - Sin( th1))
          - ( m2 / ( m1 + m2)) * Sin( th1 - th2) * (( l2 / l1) * t2d * t2d + t1d * t1d * Cos( th1 - th2))
        )
        / ( 1 - (( m2 / ( m1 + m2)) * Cos( th1 - th2) * Cos( th1 - th2)));
    end;


    function    TPendulum.th2_yder( t1d, th1, t2d, th2, l1, l2, m1, m2: TSignal): TSignal;
    begin
      Result :=
        INCREMENT * FSpeed * Jitter *
        (
          (( STANDARD_GRAVITY * FGMod) / l2) * ( Sin( th1) * Cos( th1 - th2) - Sin( th1))
          + Sin( th1 - th2) * (( l1 / l2) * t1d * t1d + ( m2 / ( m1 + m2)) * t2d * t2d * Cos( th1 - th2))
        )
        / ( 1 - (( m2 / ( m1 + m2)) * Cos( th1 - th2) * Cos( th1 - th2)));
    end;


//  public

    constructor TPendulum.Create( l1, l2, m1, m2: TSignal);
    begin
      FL1 := l1;
      FL2 := l2;
      FM1 := m1;
      FM2 := m2;
      Reset;
    end;


    procedure   TPendulum.Reset;
    begin
      FSpeed := 1.0;
      FGMod  := 1.0;
      FTh1   := 90.0 * PI / 180.0;
      FTh2   :=  0.0 * PI / 180.0;
      FT1d   :=  0.0;
      FT2d   :=  0.0;
    end;


    procedure   TPendulum.Step( out x1, y1, x2, y2, ph1, ph2: TSignal; fetch_only: Boolean);
    var
      k1, k2, k3, k4 : TSignal;
      q1, q2, q3, q4 : TSignal;
      r1, r2, r3, r4 : TSignal;
      s1, s2, s3, s4 : TSignal;
    begin
      if not fetch_only
      then begin
        k1 := th1_xder( FT1d);
        q1 := th1_yder( FT2d);
        r1 := th2_xder( FT1d, FTh1, FT2d, FTh2, FL1, FL2, FM1, FM2);
        s1 := th2_yder( FT1d, FTh1, FT2d, FTh2, FL1, FL2, FM1, FM2);

        k2 := th1_xder(  FT1d + ( r1 * 0.5));
        q2 := th1_yder(  FT2d + ( s1 * 0.5));
        r2 := th2_xder(( FT1d + ( r1 * 0.5)), ( FTh1 + ( k1 * 0.5)), ( FT2d + ( s1 * 0.5)), ( FTh2 + ( q1 * 0.5)), FL1, FL2, FM1, FM2);
        s2 := th2_yder(( FT1d + ( r1 * 0.5)), ( FTh1 + ( k1 * 0.5)), ( FT2d + ( s1 * 0.5)), ( FTh2 + ( q1 * 0.5)), FL1, FL2, FM1, FM2);

        k3 := th1_xder(  FT1d + ( r2 * 0.5));
        q3 := th1_yder(  FT2d + ( s2 * 0.5));
        r3 := th2_xder(( FT1d + ( r2 * 0.5)), ( FTh1 + ( k2 * 0.5)), ( FT2d + ( s2 * 0.5)), ( FTh2 + ( q2 * 0.5)), FL1, FL2, FM1, FM2);
        s3 := th2_yder(( FT1d + ( r2 * 0.5)), ( FTh1 + ( k2 * 0.5)), ( FT2d + ( s2 * 0.5)), ( FTh2 + ( q2 * 0.5)), FL1, FL2, FM1, FM2);

        k4 := th1_xder(  FT1d + r3);
        q4 := th1_yder(  FT2d + s3);
        r4 := th2_xder(( FT1d + r3), ( FTh1 + k3), ( FT2d + s3), ( FTh2 + q3), FL1, FL2, FM1, FM2);
        s4 := th2_yder(( FT1d + r3), ( FTh1 + k3), ( FT2d + s3), ( FTh2 + q3), FL1, FL2, FM1, FM2);

        FT1d := FT1d + ( r1 + 2.0 * r2 + 2.0 * r3 + r4) / 6.0;
        FT2d := FT2d + ( s1 + 2.0 * s2 + 2.0 * s3 + s4) / 6.0;
        FTh1 := FTh1 + ( k1 + 2.0 * k2 + 2.0 * k3 + k4) / 6.0;
        FTh2 := FTh2 + ( q1 + 2.0 * q2 + 2.0 * q3 + q4) / 6.0;
      end;

      FTh1 := MathFloatMod( FTh1, TWO_PI);
      FTh2 := MathFloatMod( FTh2, TWO_PI);

      x1  :=      FL1 * Sin( FTh1);
      y1  :=    - FL1 * Cos( FTh1);
      x2  := x1 + FL2 * Sin( FTh2);
      y2  := y1 - FL2 * Cos( FTh2);
      ph1 := ( FTh1 - PI) * PI_REC;
      ph2 := ( FTh2 - PI) * PI_REC;
    end;


end.
