unit fft_funcs;

{

  (C) COPYRIGHT 2014 .. 2020 Blue Hell / Jan Punter

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation;

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  For all listed email addresses :

    _dot. to be substituted by a dot      '.'
    2@t2  to be substituted by an at sign '@'


  Blue Hell is a trade mark owned by

    Jan Punter
    https://www.bluehell.nl/
    jan2@t2mail_dot_bluehell_dot_nl
}

{
  FFT support tools to be used in modules
}

interface

uses

  System.Math, System.SysUtils,

  KnobsUtils, ComplexMath, fftwgen, fftw_, windowing;


type

  TFFTData = record
  private
    FWindow       : TFFTWindow;
    FBinCount     : Integer;
    FIndex        : Integer;
    FFTWInBuffer  : Pfftw_real;
    FFTWOutBuffer : Pfftw_complex;
    FFTWPlan      : fftw_plan;
  public
    procedure   Initialize( aBinCount, aStartIndex: Integer; aWindow: TFFTWindow);
    procedure   Finalize;
    procedure   Tick( anInValue: TSignal; var anOutValue: TComplex);
  end;
  TFFTDataArray = array of TFFTData;


  TIFFTData = record
  private
    FWindow       : TFFTWindow;
    FBinCount     : Integer;
    FIndex        : Integer;
    FFTWInBuffer  : Pfftw_complex;
    FFTWOutBuffer : Pfftw_real;
    FFTWPlan      : fftw_plan;
  public
    procedure   Initialize( aBinCount, aStartIndex: Integer; aWindow: TFFTWindow);
    procedure   Finalize;
    procedure   Tick( anInValue: TComplex; var anOutValue: TSignal);
  end;
  TIFFTDataArray = array of TIFFTData;


  TFFTHandler = class
  private
    FDestroying : Boolean;
    FBinCount   : Integer;
    FWindow     : TFFTWindow;
    FData       : TFFTDataArray;
  private
    function    GetWindowCount: Integer;
  public
    constructor Create( aWindowCount, aBinCount: Integer; aWindowType: TFFTWindowClass);
    destructor  Destroy;                                                                                       override;
    procedure   Tick( anInValue: TSignal; var anOutValue: TComplex);
  public
    property    WindowCount: Integer read GetWindowCount;
  end;


  TIFFTHandler = class
  private
    FDestroying : Boolean;
    FBinCount   : Integer;
    FWindow     : TFFTWindow;
    FData       : TIFFTDataArray;
  private
    function    GetWindowCount: Integer;
  public
    constructor Create( aWindowCount, aBinCount: Integer; aWindowType: TFFTWindowClass);
    destructor  Destroy;                                                                                       override;
    procedure   Tick( anInValue: TComplex; var anOutValue: TSignal);
  public
    property    WindowCount: Integer read GetWindowCount;
  end;



implementation



{ ========
  TFFTData = record
  private
    FWindow       : TFFTWindow;
    FBinCount     : Integer;
    FCount        : Integer;
    FFTWInBuffer  : Pfftw_real;
    FFTWOutBuffer : Pfftw_complex;
    FFTWPlan      : fftw_plan;
  end;
  TFFTDataArray = array of TFFTData;
}

//  public

    procedure   TFFTData.Initialize( aBinCount, aStartIndex: Integer; aWindow: TFFTWindow);
    var
      i : Integer;
    begin
      FWindow       := aWindow;
      FBinCount     := aBinCount;
      FIndex        := aStartIndex;
      FFTWInBuffer  := fftw_malloc( aBinCount * SizeOf( Tfftw_real   ));
      FFTWOutBuffer := fftw_malloc( aBinCount * SizeOf( Tfftw_complex));
      FFTWPlan      := fftw_plan_dft_r2c_1d( aBinCount, FFTWInBuffer, FFTWOutBuffer, FFTW_MEASURE);

      for i := 0 to aBinCount - 1
      do begin
        Pfftw_complex_array( FFTWOutBuffer)^[ i].re.v := 0;
        Pfftw_complex_array( FFTWOutBuffer)^[ i].im.v := 0;
        Pfftw_real_array   ( FFTWInBuffer )^[ i].v    := 0;
      end;
    end;


    procedure   TFFTData.Finalize;
    begin
      if   Assigned( FFTWInBuffer)
      then fftw_free( FFTWInBuffer);

      if   Assigned( FFTWOutBuffer)
      then fftw_free( FFTWOutBuffer);

      if   Assigned( FFTWPlan)
      then fftw_destroy_plan( FFTWPlan);
    end;


    procedure   TFFTData.Tick( anInValue: TSignal; var anOutValue: TComplex);
    begin
      if   FIndex >= FBinCount
      then begin
        FIndex := 0;
        fftw_execute( FFTWPlan);
      end;

      if   FIndex >= 0
      then begin
        Pfftw_real_array( FFTWInBuffer)[ FIndex].v := anInValue * FWindow[ FIndex];
        anOutValue.Real := anOutValue.Real + Pfftw_complex_array( FFTWOutBuffer)^[ FIndex].re.v;
        anOutValue.Imag := anOutValue.Imag + Pfftw_complex_array( FFTWOutBuffer)^[ FIndex].im.v;
      end;

      Inc( FIndex);
    end;


{ ========
  TIFFTData = record
  private
    FWindow       : TFFTWindow;
    FBinCount     : Integer;
    FCount        : Integer;
    FFTWInBuffer  : Pfftw_complex;
    FFTWOutBuffer : Pfftw_real;
    FFTWPlan      : fftw_plan;
  end;
  TIFFTDataArray = array of TIFFTData;
}

//  public

    procedure   TIFFTData.Initialize( aBinCount, aStartIndex: Integer; aWindow: TFFTWindow);
    var
      i : Integer;
    begin
      FWindow       := aWindow;
      FBinCount     := aBinCount;
      FIndex        := aStartIndex;
      FFTWInBuffer  := fftw_malloc( aBinCount * SizeOf( Tfftw_complex));
      FFTWOutBuffer := fftw_malloc( aBinCount * SizeOf( Tfftw_real   ));
      FFTWPlan      := fftw_plan_dft_c2r_1d( aBinCount, FFTWInBuffer, FFTWOutBuffer, FFTW_MEASURE);

      for i := 0 to aBinCount - 1
      do begin
        Pfftw_complex_array( FFTWInBuffer )^[ i].re.v := 0;
        Pfftw_complex_array( FFTWInBuffer )^[ i].im.v := 0;
        Pfftw_real_array   ( FFTWOutBuffer)^[ i].v    := 0;
      end;
    end;


    procedure   TIFFTData.Finalize;
    begin
      if   Assigned( FFTWInBuffer)
      then fftw_free( FFTWInBuffer);

      if   Assigned( FFTWOutBuffer)
      then fftw_free( FFTWOutBuffer);

      if   Assigned( FFTWPlan)
      then fftw_destroy_plan( FFTWPlan);
    end;


    procedure   TIFFTData.Tick( anInValue: TComplex; var anOutValue: TSignal);
    begin
      if   FIndex >= FBinCount
      then begin
        FIndex := 0;
        fftw_execute( FFTWPlan);
      end;

      if   FIndex >= 0
      then begin
        Pfftw_complex_array( FFTWInBuffer)[ FIndex].re.v := anInValue.Real;
        Pfftw_complex_array( FFTWInBuffer)[ FIndex].im.v := anInValue.Imag;
        anOutValue := anOutValue + FWindow[ FIndex] * Pfftw_real_array( FFTWOutBuffer)^[ FIndex].v;
      end;

      Inc( FIndex);
    end;


{ ========
  TFFTHandler = class
  private
    FBinCount : Integer;
    FWindow   : TFFTWindow;
    FData     : TFFTDataArray;
  public
    property    WindowCount: Integer read GetWindowCount;
  private
}

    function    TFFTHandler.GetWindowCount: Integer;
    begin
      Result := Length( FData);
    end;


//  public

    constructor TFFTHandler.Create( aWindowCount, aBinCount: Integer; aWindowType: TFFTWindowClass);
    var
      i : Integer;
    begin
      Assert( aBinCount    > 0);
      Assert( aWindowCount > 0);
      inherited Create;
      FDestroying := False;
      FBinCount   := aBinCount;
      FWindow     := aWindowType.Create( aBinCount);
      SetLength( FData, aWindowCount);

      for i := 0 to aWindowCount - 1
      do FData[ i].Initialize( FBinCount, - (( i * FBinCount) div aWindowCount), FWindow);
    end;


    destructor  TFFTHandler.Destroy; // override;
    begin
      FDestroying := True;

      while WindowCount > 0
      do begin
        FData[ WindowCount - 1].Finalize;
        SetLength( FData, WindowCount - 1);
      end;

      FreeAndNil( FWindow);
      inherited;
    end;


    procedure   TFFTHandler.Tick( anInValue: TSignal; var anOutValue: TComplex);
    var
      i : Integer;
    begin
      anOutValue.Real := 0;
      anOutValue.Imag := 0;

      if   not FDestroying
      then begin
        for i := 0 to WindowCount - 1
        do FData[ i].Tick( anInValue, anOutValue);

        anOutValue := anOutValue / WindowCount;
      end;
    end;


{ ========
  TIFFTHandler = class
  private
    FBinCount : Integer;
    FWindow   : TFFTWindow;
    FData     : TIFFTDataArray;
  public
    property    WindowCount: Integer read GetWindowCount;
  private
}

    function    TIFFTHandler.GetWindowCount: Integer;
    begin
      Result := Length( FData);
    end;


//  public

    constructor TIFFTHandler.Create( aWindowCount, aBinCount: Integer; aWindowType: TFFTWindowClass);
    var
      i : Integer;
    begin
      Assert( aBinCount    > 0);
      Assert( aWindowCount > 0);
      inherited Create;
      FDestroying := False;
      FBinCount   := aBinCount;
      FWindow     := aWindowType.Create( aBinCount);
      SetLength( FData, aWindowCount);

      for i := 0 to aWindowCount - 1
      do FData[ i].Initialize( FBinCount, ( i * FBinCount) div aWindowCount, FWindow);
    end;


    destructor  TIFFTHandler.Destroy; // override;
    begin
      FDestroying := True;

      while WindowCount > 0
      do begin
        FData[ WindowCount - 1].Finalize;
        SetLength( FData, WindowCount - 1);
      end;

      FreeAndNil( FWindow);
      inherited;
    end;


    procedure   TIFFTHandler.Tick( anInValue: TComplex; var anOutValue: TSignal);
    var
      i : Integer;
    begin
      anOutValue := 0;

      if   not FDestroying
      then begin
        for i := 0 to WindowCount - 1
        do FData[ i].Tick( anInValue, anOutValue);

        anOutValue  := anOutValue  / WindowCount;
      end;
    end;


end.

